;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  File: statistics.lisp
;;;  Author: Chichilnisky
;;;  Description: Simple statistical routines. Various functions for generating
;;;               noise from distributions, other functions associated with data analysis.
;;;  Creation Date: 5/23/92
;;;  ----------------------------------------------------------------
;;;    Object-Based Vision and Image Understanding System (OBVIUS),
;;;      Copyright 1988, Vision Science Group,  Media Laboratory,  
;;;              Massachusetts Institute of Technology.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(in-package 'obvius)

;; General statistics routines.
(export '(check-probability check-integer
          weibull inverse-weibull
	  detection-weibull inverse-detection-weibull 2afc-weibull inverse-2afc-weibull
          detection-log-normal inverse-detection-log-normal
          chi-square cumulative-chi-square cumulative-f cumulative-t
          poisson cumulative-poisson poisson-noise factorial
          cumulative-binomial binomial binomial-noise bernoulli-noise
	  binomial-coefficient log-binomial-coefficient
	  standard-deviation sample-variance sample-standard-deviation standard-error
          confidence-interval mean-confidence-interval))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Common error-checking for stats functions.
(defun check-probability (p)
  (unless (<= 0.0 p 1.0)
    (error "~a is not in the range (0,1)" p)))

(defun check-integer (&rest numbers)
  (dolist (number numbers)
    (unless (= (ffloor number) (float number))
      (error "~a is not an integer" number))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Statistics on finite samples
(defun standard-deviation (thing)
  (sqrt (variance thing)))

(defun sample-variance (thing)
  (* (variance thing) (/ (total-size thing) (- (total-size thing) 1))))

(defun sample-standard-deviation (thing)
  (sqrt (sample-variance thing)))

(defun standard-error (thing)
  (/ (sample-standard-deviation thing) (sqrt (total-size thing))))

(defun confidence-interval (thing)
  (let ((mean (mean thing))
        (range (* 2.0 (sample-standard-deviation thing))))
    (list (- mean range) (+ mean range))))

(defun mean-confidence-interval (thing)
  (let ((mean (mean thing))
        (range (* 2.0 (standard-error thing))))
    (list (- mean range) (+ mean range))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Weibull
(defun weibull (x alpha beta gamma)
  (- 1.0 (* (- 1.0 gamma) (exp (- (expt (/ x alpha) beta))))))

(defun inverse-weibull (w alpha beta gamma)
  (* alpha (expt (- (log (/ (- 1.0 w) (- 1.0 gamma)))) (/ 1.0 beta))))

(defun detection-weibull (x alpha beta)
  (- 1.0 (exp (- (expt (/ x alpha) beta)))))

(defun inverse-detection-weibull (w alpha beta)
  (* alpha (expt (- (log (- 1.0 w))) (/ 1.0 beta))))

(defun 2afc-weibull (x alpha beta)
  (- 1.0 (* 0.5 (exp (- (expt (/ x alpha) beta))))))

(defun inverse-2afc-weibull (w alpha beta)
  (* alpha (expt (- (log (/ (- 1.0 w) 0.5))) (/ 1.0 beta))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Cumulative normal on a log scale

(defun detection-log-normal (x half slope)
  (cumulative-normal (* slope (log (/ x half) 2.0))))

(defun inverse-detection-log-normal (x half slope)
  (* half (expt 2.0 (/ (inverse-cumulative-normal x) slope))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Chi-square
(defun cumulative-chi-square (x df)
  (check-integer df)
  (assert (plusp df))
  (gammp (/ df 2.0) (/ x 2.0)))
#|
;; should be 0.95
(cumulative-chi-square 11.07 5)
(cumulative-chi-square 18.31 10)

;; should be 0.05
(cumulative-chi-square 1.15 5)
(cumulative-chi-square 3.94 10)

|#

(defun chi-square (x df &key (sigma 1.0))
  (check-integer df)
  (let ((df/2 (/ df 2.0)))
    (if (plusp x)
	(* (exp (- (/ x (* 2.0 (sqr sigma)))))
	   (expt x (- df/2 1))
	   (/ (expt 2.0 df/2))
	   (/ (expt sigma df))
	   (/ (exp (gammln df/2))))	; exp of gammln is gamma function
	0.0)))
#|
(make-discrete-function #'(lambda(x) (chi-square x 2)) 0.0 10.0 :size 128)
|#

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Poisson
(defun poisson (k lambda)
  (/ (* (expt lambda k) (exp (- lambda))) (factorial k)))

#|
(make-discrete-function #'(lambda(x) (poisson x 2)) 0.0 10.0 :size 11)
(distribution-mean *)

(make-discrete-function #'(lambda(x) (poisson x 10)) 0.0 50.0 :size 51)
(distribution-mean *)
|#

;; The numerical recipes version works up to 35, then overflows single-floats.
;; Lisp works beyond that, but is slow. So, we use a hybrid.
(defun factorial (x)
  (check-integer x)
  (cond ((minusp x) (error "factorial not defined for negative arguments ~a" x))
	((zerop x) 1)
	((> x 34) (* x (factorial (- x 1))))
	(t (floor (factrl (floor x))))))

#|
(factorial 2)
(factorial 3)
(factorial 4)
(factorial 5)
(factorial 6)
|#

;;; Pr{observed < count}
(defun cumulative-poisson (count lambda)
  (check-integer count)
  (if (zerop count) 0.0
      (gammq (ffloor count) (float lambda))))

#|
;; Test poisson distribution and cumulative
(let ((lambda 100)
      (count 80))
  (print (loop for i from 0 below count
	       sum (poisson i lambda)))
  (print (cumulative-poisson count lambda))
  nil)

|#

;;; Two ways to compute Poisson noise:
;;; 1) Use a NR approximation to the CDF
;;; 2) Build up the CDF as you go along
;;; The former's run-time is too slow for low-mean Poisson.
;;; The latter's run-time climbs with mean of Poisson, so bad for high mean.
;;; So, use hybrid.
;;; Note: (exp) underflows around -700. This is a way to set the cutoff.
;;; *** If the probabilities were done with logs, thing might work better for high mean.
(defun poisson-noise (lambda)
  (if (< lambda 700)
      ;; Build up the CDF as you go
      (let ((random (random 1.0))
	    (flambda (float lambda))	; Save floating conversions
	    (probability (exp (- lambda))) ; Current probability
	    (cumulative (exp (- lambda)) ) ; Cumulative probability
	    )
	(declare (single-float random cumulative probability flambda))
	(when (zerop probability)
	  (error "Poisson exponent too large"))
	(loop for i from 1
	      when (> cumulative random)
	      return (1- i)
	      do
	      (setq probability (* probability (/ flambda (float i))))
	      (incf cumulative probability)))
      ;; Use NR approximation to CDF
      (loop for i from 0
	    with random = (random 1.0)
	    when (> (cumulative-poisson i lambda) random)
	    return (1- i))))

#|
;;; Test Poisson noise
(length (setf vec (make-array 1000 :element-type 'single-float)))
(progn
  (dotimes (i (length vec))
    (setf (aref vec i) (float (poisson-noise 6))))
  (values (mean vec) (variance vec)))

|#

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; F distribution, with two degrees of freedom
(defun cumulative-f (x df-1 df-2)
  (check-integer df-1 df-2)
  (if (plusp x)
      (- 1.0 (betai (/ df-2 2.0) (/ df-1 2.0) (/ (float df-2) (+ df-2 (* df-1 x)))))
      0.0))

#|
;; should all be 0.90
(cumulative-f 55.83 4 1)
(cumulative-f 4.19 3 4)

;; should all be 0.95
(cumulative-f 9.28 3 3)
(cumulative-f 3.18 9 9)
|#

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; t distribution
(defun cumulative-t (x df)
  (check-integer df)
  (let ((result (- 1.0 (/ (betai (/ df 2.0) 0.5 (/ df (+ df (sqr x)))) 2.0))))
    (if (plusp x) result (- 1.0 result))))
      

#|
;; should all be 0.95
(cumulative-t 6.314 1)
(cumulative-t 1.943 6)
(cumulative-t 1.796 11)

;; should all be 0.6
(cumulative-t 0.325 1)
(cumulative-t 0.265 6)
(cumulative-t 0.256 26)

|#

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Binomial distribution and Bernoulli process.

;; This is the probability of n or fewer successes (not probability of n or more).
;; This is consistent with cumulative distributions which rise up to 1,
;; rather than the NR terminology. 
(defun cumulative-binomial (count p trials)
  (check-integer trials count)
  (let ((limit (+ count 1)))
    (cond ((>= limit trials) 1.0)
	  ((minusp limit) 0.0)
	  ((- 1.0 (betai (float limit) (+ 1.0 (- trials limit)) (float p)))))))

;;; Probability of k successes from a binomial draw with probability p and n trials
(defun binomial (k p n)
  (check-probability p)
  (* (expt p k) (expt (- 1.0 p) (- n k)) (binomial-coefficient n k)))

#|
;; Test poisson distribution and cumulative
(let ((trials 20)
      (count 10)
      (p 0.5))
  (print (loop for i from 0 upto count
	       sum (binomial i p trials)))
  (print (cumulative-binomial count p trials))
  nil)

|#

;;; Number of distinct ways of choosing k binomial successes in n trials.
;;; Alternately, number of distinct binary strings of length n that contain k 1s.
;;; Same as the words "n choose k"
;;; Use the Numerical Recipes version for speed
(defun binomial-coefficient (n k)
  (check-integer n k)
  (round (bico (floor n) (floor k))))

(defun log-binomial-coefficient (n k)
  (check-integer n k)
  (- (factln n) (factln k) (factln (- n k))))

#|
(let ((n 5)
      (k 2))
  (print (binomial-coefficient n k))
  (print (/ (factorial n) (* (factorial k) (factorial (- n k)))))
  nil)
|#

#|
;; Lisp version is slower than NR version.
(defun binomial-coefficient (n k)
  (loop with result = 1
	for index from n above k
	do (setq result (* result index))
	finally return (/ result (factorial (- n k)))))
|#

(defun binomial-noise (p &optional (trials 1))
  (check-integer trials)
  (loop for i from 0 below trials count (not (zerop (bernoulli-noise p)))))

(defun bernoulli-noise (p)
  (if (< (random 1.0) p) 1 0))

;;; Local Variables:
;;; buffer-read-only: t 
;;; End:
